#include <stdio.h>
#include <stdlib.h>

#include "huffman.h"

int is_leaf(const struct tree *t)
{
    return t->left == NULL && t->right == NULL;
}

struct tree *make_tree(int value, int letter, struct tree *left, struct tree *right)
{
    struct tree *l = malloc(sizeof(*l));
    l->value = value;
    l->letter = letter;
    l->left = left;
    l->right = right;
    return l;
}

void free_tree(struct tree *root)
{
    if (root->left)
        free_tree(root->left);
    if (root->right)
        free_tree(root->right);
    root->left = root->right = NULL;
    free(root);
};

static struct tree *make_leaf(int value, unsigned char letter)
{
    return make_tree(value, letter, NULL, NULL);
}

static struct tree *make_node(struct tree *left, struct tree *right)
{
    return make_tree(left->value + right->value, 0, left, right);
}

// TODO
int rank(const struct tree *node)
;

// TODO
const struct tree *find_right_most_leaf(const struct tree *node)
;

/**
 * Fill `array` (of `size`) with `value`
 */
static void fill_int_array(int value, int size, int array[])
{
    for (int i = 0; i < size; i ++)
        array[i] = value;
}

int hist(int length, const unsigned char data[], int result[])
{
    int k = 0;
    fill_int_array(0, MAX_UCHAR, result);
    for (int i = 0; i < length; i++) {
        if (result[data[i]] ++ == 0)
            k ++;
    }
    return k;
}

// TODO
int hist_stream(FILE *stream, int result[])
;

/**
 * Create a forest (i.e., an array of tree) associating each possible byte
 * to its frequency.
 * Returns the number of letters associated this way.
 */
static int make_letters(const int freq[], struct tree *letters[])
{
    int k = 0;
    for (int i = 0; i < MAX_UCHAR; i ++) {
        if (freq[i])
            letters[k++] = make_leaf(freq[i], i);
    }
    for (int i = k; i < MAX_UCHAR; i++)
        letters[i] = NULL;
    return k;
}

/**
 * Find the index of the tree with the lowest frequency
 * Prefer tree over letters
 */
static int index_min(int size, struct tree *letters[])
{
    int idx = 0;
    for (int i = 1; i < size; i ++) {
        if (letters[i]->value < letters[idx]->value
                || (letters[i]->value == letters[idx]->value
                    && letters[i]->value > letters[idx]->value))
            idx = i;
    }
    return idx;
}

struct tree *make_huffman_tree(int freq[])
{
    struct tree *letters[MAX_UCHAR] = {};
    int size = make_letters(freq, letters);

    while (size > 1) {
        int idx = index_min(size, letters);
        struct tree *min_tree = letters[idx];
        letters[idx] = letters[--size];

        idx = index_min(size, letters);
        struct tree *other_min_tree = letters[idx];
        letters[idx] = letters[size - 1];

        if (rank(other_min_tree) < rank(min_tree))
            letters[size - 1] = make_node(other_min_tree, min_tree);
        else
            letters[size - 1] = make_node(min_tree, other_min_tree);
        letters[size - 1]->letter =
            other_min_tree->letter > min_tree->letter ? other_min_tree->letter : min_tree->letter;
    }
    return letters[0];
}

static void make_compress_table_rec(struct bit_buffer prefix, const struct tree *root, struct bit_buffer table[])
{
    if (is_leaf(root)) {
        table[root->letter] = prefix;
    } else {
        prefix.count ++;
        prefix.bits <<= 1;
        if (root->left)
            make_compress_table_rec(prefix, root->left, table);
        prefix.bits |= 1;
        if (root->right)
            make_compress_table_rec(prefix, root->right, table);
    }
}

static void patch_compress_table(const struct tree *root, struct bit_buffer table[])
{
    const struct tree *right_most_leaf = find_right_most_leaf(root);
    struct bit_buffer bb = table[right_most_leaf->letter];
    bb.count ++;
    bb.bits <<= 1;
    table[right_most_leaf->letter] = bb;
    bb.bits |= 1;
    table[EOS] = bb;
}

void make_compress_table(const struct tree *root, struct bit_buffer table[])
{
    struct bit_buffer prefix = {};
    make_compress_table_rec(prefix, root, table);
}

/**
 * Return remaining space of `buffer`
 */
static int bitbuffer_remaining(const struct bit_buffer *buffer)
{// Given
    return BIT_BUFFER_SIZE - buffer->count;
}

/**
 * Return non zero if `buffer` is empty
 */
static int bitbuffer_empty(const struct bit_buffer *buffer)
{// Given
    return buffer->count == 0;
}

/**
 * Return non zero if `buffer` is full
 */
static int bitbuffer_full(const struct bit_buffer *buffer)
{// Given
    return buffer->count == BIT_BUFFER_SIZE;
}

/**
 * Transfer `count` most significant bit from `data` to less significant bit
 * of `buffer`.  `data` should contain at least `count` bits, and `buffer`
 * should have at least `count` bits remaining
 */
static void transfer_to_buffer(int count, struct bit_buffer *buffer, struct bit_buffer *data)
{// Given
    buffer->bits <<= count;
    buffer->bits |= (data->bits >> (data->count - count)) & N_ONES(count);
    data->count -= count;
    buffer->count += count;
}

// TODO
int read_bit(struct compressed_stream *in)
;

// TODO
int read_bits(struct compressed_stream *in, int count)
;

/**
 * Write a full `buffer` to `fd` and reset it
 * Assume that buffer is full;
 */
static int commit_buffer(FILE *fd, struct bit_buffer *buffer)
{
    buffer->count = 0;
    return fputc(buffer->bits, fd);
}

int write_bit(struct compressed_stream *out, int bit)
{
    out->buffer.bits = (out->buffer.bits << 1)  | bit;
    out->buffer.count ++;
    if (bitbuffer_full(&out->buffer))
        if (commit_buffer(out->stream, &out->buffer) == EOF)
            return EOF;
    return bit;
}

int write_bits(struct compressed_stream *out, int count, int bits)
;

void compress_byte(struct compressed_stream *out, unsigned char byte)
{// Given
    struct bit_buffer data = out->compress_table[byte];
    do {
        int remaining = bitbuffer_remaining(&out->buffer);
        int count = (data.count < remaining) ? data.count : remaining;
        transfer_to_buffer(count, &out->buffer, &data);
        if (bitbuffer_full(&out->buffer))
            commit_buffer(out->stream, &out->buffer);
    } while (!bitbuffer_empty(&data));
}

struct compressed_stream compress_stream(FILE *stream, struct tree *root)
{
    struct compressed_stream out = {
        .stream = stream,
        .root = NULL,
        .compress_table = malloc(sizeof(struct bit_buffer)*(MAX_UCHAR + 1)),
    };

    write_tree(&out, root);
    // print_tree_to_dot(stderr, root);

    make_compress_table(root, out.compress_table);
    patch_compress_table(root, out.compress_table);
    // print_compress_table(stderr, out.compress_table);

    return out;
}

void compress(struct compressed_stream *out, FILE *in)
{
    int byte;
    while ((byte = fgetc(in)) != EOF)
        compress_byte(out, byte);
}

void write_compressed(struct compressed_stream *out, int len, const char data[])
{
    for (int i = 0; i < len; i++)
        compress_byte(out, data[i]);
}

void flush_compressed(struct compressed_stream *out)
{
    if (out->compress_table) {
        write_bits(out, out->compress_table[EOS].count, out->compress_table[EOS].bits);
        //write_bits(out, bitbuffer_remaining(&out->buffer), -1); // flood with one
        write_bits(out, bitbuffer_remaining(&out->buffer), 0); // flood with zero
    }
    fflush(out->stream);
}

void close_compressed(struct compressed_stream *out)
{
    flush_compressed(out);

    if (out->compress_table)
        free(out->compress_table);
    out->compress_table = NULL;

    if (out->root)
        free_tree(out->root);
    out->root = NULL;
    out->stream = NULL;
}

static struct tree *add_eos(struct tree *root)
{
    struct tree *right_most_leaf = (struct tree*)find_right_most_leaf(root);
    right_most_leaf->left = make_leaf(0, right_most_leaf->letter);
    right_most_leaf->right = make_tree(0, EOS, NULL, NULL);
    return root;
}

struct compressed_stream uncompress_stream(FILE *stream)
{
    struct compressed_stream in = {
        .stream = stream,
    };
    struct tree *root = read_tree(&in);
    if (root == NULL)
        in.stream = NULL;
    else {
        in.root = add_eos(root);
    }
    return in;
}

// TODO
int decompress_byte(struct compressed_stream *in)
;

// TODO
int read_compressed(struct compressed_stream *in, int len, char data[])
;

/**
 * Write the tree `node` in `stream`
 * Return number of tree written or 0 if error.
 *
 * Algorithm:
 * - Leaf: write `1`, then the 8 bits of symbol.
 * - Node: Write `0`, and recursively write left then right.
 */
//TODO
int write_tree(struct compressed_stream *stream, const struct tree *node)
;

/**
 * Associe le symbole de la feuille `leaf` aux 8 prochains bits.
 * Retourne `leaf` ou NULL si erreur de lecture.
 */
//TODO
static struct tree *read_leaf(struct compressed_stream *stream, struct tree *leaf)
;

/**
 * Remplit l'arbre `node` en fonction du flot `stream`.
 *
 * En fonction du premier bit, détermine si c'est une feuille (1), ou un noeud
 * interne (0).
 */
//TODO
static struct tree *read_node(struct compressed_stream *stream, struct tree *node)
;

/* TODO Uncomment when you try to make read_node/read_leaf

struct tree *read_tree(struct compressed_stream *stream)
{
    struct tree *node = make_tree(0, 0, NULL, NULL);
    if (read_node(stream, node) == NULL)
        free_tree(node);
    return node;
}

*/
