#include <stdio.h>

#define MAX_UCHAR       (1<<8)
#define N_ONES(__n)     ((1 << (__n)) - 1)

#define EOS             MAX_UCHAR

struct tree {
    int value;
    int letter; // Valid only if this tree is a leaf

    struct tree *left;
    struct tree *right;
};


// Currently this is a constant, it should be a member of `struct bit_buffer`
#define BIT_BUFFER_SIZE (sizeof(unsigned char)*8)

struct bit_buffer {
    int count;
    int bits;
};

struct compressed_stream {
    FILE *stream; // The underlying (un)compressed stream
    struct bit_buffer buffer;

    struct tree *root; // The tree (only for uncompress)
    struct bit_buffer *compress_table; // The compress table (only for compress)
};

/**
 * Return non zero if `t` is a leaf
 */
int is_leaf(const struct tree *t);

/**
 * Allocate a tree with all data members
 */
struct tree *make_tree(int value, int letter, struct tree *left, struct tree *right);

/**
 * Recursively deallocate a tree.
 */
void free_tree(struct tree *root) ;

/**
 * Count occurences of `length` element of `data` and store the result
 * in `result`
 *
 * RQ: Rewrite this function to compute that for a stream
 */
int hist(int length, const unsigned char data[], int result[]);

/**
 * Count occurences of each chararcter of the stream,
 * and store the result in `result` (where characters value are the index)
 */
int hist_stream(FILE *in, int result[]);

/**
 * Create an huffman tree with `freq` probabilities.
 * It's assumed that len(array) == MAX_UCHAR
 *
 * Algorithm:
 *   - Create a forest of (letters, frequency)
 *   - Until letter is a single tree
 *      - Pick the tree with the lowest frequency
 *      - Exchange it with the last one
 *      - Reduce size of 1
 *      - Pick the tree with the lowest frequency except the previous one
 *      - Replace it by a new tree parent of the two chosen one
 */
struct tree *make_huffman_tree(int freq[]);

/**
 * Associate in `table` a bit_buffer for each leaf of `root`.
 * Left branchis are 0, Right are 1.
 */
void make_compress_table(const struct tree *root, struct bit_buffer table[]);

/**
 * Dump a tree `root` to the Graphviz dot format.
 */
void print_tree_to_dot(FILE *out, struct tree *root);

/**
 * Dump on `out` a compress table
 */
void print_compress_table(FILE *out, const struct bit_buffer table[]);

/**
 * Associate `stream` to an huffman tree.
 *
 * Allocate & compute the compress table.
 */
struct compressed_stream compress_stream(FILE *stream, struct tree *root);

/**
 * Write `len` `data` to the stream.
 *
 * TODO: Another function that compress another stream instead
 */
void write_compressed(struct compressed_stream *out, int len, const char data[]);

/**
 * Compress the stream `in` into a compressed stream `out`
 */
void compress(struct compressed_stream *out, FILE *in);

/**
 * Add `byte` compressed stream `out`
 *
 * RQ: This is a picky function
 */
void compress_byte(struct compressed_stream *out, unsigned char byte);

/**
 * Finalize the compress stream by flushing whatever left.
 */
void flush_compressed(struct compressed_stream *out);

/**
 * Close this compressed stream.
 * Ensure that everything is writen, and deal with ressource deallocation.
 */
void close_compressed(struct compressed_stream *out);


/**
 * Associate the input `stream` to a compressed stream.
 * This stream should have be created by `compressed_stream`
 */
struct compressed_stream uncompress_stream(FILE *stream);

/**
 * Read one symbol from `in` and return it, or return EOF if the stream ends.
 */
int decompress_byte(struct compressed_stream *in);

/**
 * Read at most `len` symbols from `in` and store them into `data`.
 * Returns the number of symbol read.
 */
int read_compressed(struct compressed_stream *in, int len, char data[]);


// *** All of theses functions ought to be static, for the shake of this
// exercice, they are  listead instead here ***

/**
 * Write the tree `node` in `stream`
 * Return number of tree written or 0 if error.
 *
 * Algorithm:
 * - Leaf: write `1`, then the 8 bits of symbol.
 * - Node: Write `0`, and recursively write left then right.
 */
int write_tree(struct compressed_stream *stream, const struct tree *root);

/**
 * Reads a tree from `stream`, return NULL on error.
 */
struct tree *read_tree(struct compressed_stream *stream);

/**
 * print an error and exit if condition is true (non zero)
 */
void exit_if(int condition, const char *prefix);

/**
 * These functions ought to be static. However they can't since an
 * implementation is provided.
 */

/**
 * Return one bit from `in`
 */
int read_bit(struct compressed_stream *in);

/**
 * Return `count` bits from `in` in bigendian.
 * It return EOF if not all of them may be read.
 */
int read_bits(struct compressed_stream *in, int count);

/**
 * Wrtite one bit from `in`
 */
int write_bit(struct compressed_stream *out, int bit);

/**
 * Write the `count` less significative bits of `bits` to `out` in bigendian.
 * It return the number of bits written or EOF
 */
int write_bits(struct compressed_stream *out, int count, int bits);

/**
 * Return the rank of a node.
 *
 * Algorithm:
 * - Rank of a leaf : 1
 * - Rank of an internal node : max rank of any child + 1
 */
int rank(const struct tree *node);

/**
 * Return the right most leaf of a tree `node`.
 */
const struct tree *find_right_most_leaf(const struct tree *node);
